#!/usr/bin/env python3
from argparse import Namespace
from base64 import b64decode
from fnmatch import fnmatch
from json import loads, JSONDecodeError
from typing import List

from boto3 import client


SECRETS_MANAGER = client("secretsmanager")


def get_secret_list(patterns: List[str], client: object = SECRETS_MANAGER) -> dict:
    """
    Generate a list of secrets, organized by the pattern that matched them. A dict containing
    the input patters as the key and a list of matched secret names as the value is returned
    """
    opts = {"MaxResults": 100}
    matches = {}
    secrets_list = []
    parsed_names = []

    # Get all of the secrets in the account
    while True:
        res = client.list_secrets(**opts)
        secrets_list += [x["Name"] for x in res["SecretList"]]
        if token := res.get("NextToken"):
            opts["NextToken"] = token
        else:
            break

    for pattern in patterns:
        # If it's just a name and not a glob match it and move on
        if "*" not in pattern and pattern in res:
            matches[pattern] = [pattern]
            parsed_names.append(pattern)
            continue
        # Get all secret names that match the current pattern
        matches[pattern] = [
            x for x in secrets_list if x not in parsed_names and fnmatch(x, pattern)
        ]
        # Append to the list of parsed names so each secret is only matched once
        parsed_names += [x for x in secrets_list if x in matches[pattern]]

    return matches


def make_secret_name(name: str, config: dict) -> str:
    if config["upper"]:
        name = name.upper()
    if config["lower"]:
        name = name.lower()
    if prefix := config.get("prefix"):
        name = f"{prefix}{name}"
    return name


def handle_secret(secret: dict) -> dict:
    vals = secret.get("SecretString", "SecretBinary")

    # Secret values could be in different formats depending on how they were stored. See the response
    # section of the boto3 docs for `get_secret_value()` for specifics:
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/secretsmanager/client/get_secret_value.html
    if isinstance(vals, bytes):
        vals = b64decode(vals).decode()

    try:
        vals = loads(vals)
    except (TypeError, JSONDecodeError):
        pass

    # If the value wasn't a JSON object then set the name to the name of the secret
    # itself since there will only be one value and no nested keys
    if not isinstance(vals, dict):
        vals = {secret["Name"]: vals}

    return vals


def get_secrets(args: Namespace | List[Namespace], client = SECRETS_MANAGER):
    """
    Retreive all of the scret values for the list provided by get_secrets_list()
    """
    configs = {}

    if not isinstance(args, list):
        args = [args]

    # This value would always be the same for all entries, regardless of
    # whether it was commandline args or stdin
    merge = args[0].merge_keys

    # Create a config that allows us to match up patters to their original
    # commandline arguments and store the values from any matching secret names
    for arg in args:
        # Terminal garbage from reading stdin
        if len(arg.secret) == 1 and arg.secret[0] in ("-", ""):
            continue

        configs.update(
            {
                secret: {
                    "prefix": arg.prefix,
                    "upper": arg.upper,
                    "lower": arg.lower,
                    "values": {},
                }
                for secret in arg.secret
            }
        )

    patterns = list(configs.keys())

    secret_list = get_secret_list(patterns, client)

    # Take the patterns and their corresponding matched secret names and retrieve the values
    for pattern, matches in secret_list.items():
        if not matches:
            del configs[pattern]
            continue

        for match in matches:
            try:
                res = client.get_secret_value(SecretId=match)
            except client.exceptions.ResourceNotFoundException:
                res = {"SecretString": None}

            vals = handle_secret(res)

            # If the value wasn't a JSON object then set the name to the name of the secret
            # itself since there will only be one value and no nested keys
            if not isinstance(vals, dict):
                vals = {match: vals}

            # If --merge-keys wasn't set then we will retain the name of the secret as the key.
            # this means the user will have to handle parsing any items out of the secret itself.
            if not merge:
                name = make_secret_name(match, configs[pattern])
                configs[pattern]["values"][name] = vals
                continue
            # If we made it here then the secret is a JSON object and the values dict will contain
            # the keys inside of the secret. In the case of multiple secrets containing the same keys
            # the last one in wins the conflict
            for name, val in vals.items():
                name = make_secret_name(name, configs[pattern])
                configs[pattern]["values"][name] = val

    res = {}

    # Flatten all of the different values dicts. Just like when parsing the values from
    # multiple secrets with the same pattern above, the last key found will win in any conflicts
    for items in configs.values():
        res.update(items["values"])

    return res



#!/usr/bin/env python3
from argparse import ArgumentParser, RawTextHelpFormatter, Namespace
from json import dumps
from re import compile
from sys import argv, stdin, stderr
from textwrap import dedent
from typing import List

from boto3 import client
from yatc import TermColor

from .secretsmanager import get_secrets


SECRETS = client("secretsmanager")
ANSI_RE = compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")

stdin_help_string = dedent(f"""
    When using stdin, a subset of arguments are taken as lowercase strings matching the long version of argument names and separated by a pipe. You don't actually {TermColor.decorate("have", decorations=["ITALIC", "WHITE"])} to use this argument to use stdin.
    Technically just passing data on stdin is enough.
    Example:\n
    {TermColor.decorate("echo 'myvar | upper | prefix TF_VAR_;' | secrets-env -", ["WHITE", "BOLD"])}\n
    This is the equivalent of passing {TermColor.decorate("-u -p TF_VAR_ myvar", ["WHITE", "BOLD"])} as commandline arguments. {TermColor.decorate("Note the ending '-'.", ["BOLD", "ITALIC", "UNDERLINE"])} Multiple secrets are supported, with each
    one taking its own set of arguments and separated by a semi-colon.
    """
)

main_description_string = dedent(
    """
    Return one or more secrets by either name or pattern, using Unix style glob matching. When using glob style matches then all secret names that
    match the pattern will be returned. When using stdin and specifying options per pattern the options will be applied to the first match whereas multiple secrets
    supplied as positional args will have the options provided as commandline arguments applied to all of them.
    """)

def parse_args(args: list = []):

    parser = ArgumentParser(
        formatter_class=RawTextHelpFormatter,
        description=main_description_string
    )
    parser.add_argument(
        "-p", "--prefix", type=str, help="Prepend a prefix to each value", default=""
    )
    parser.add_argument(
        "-m",
        "--merge-keys",
        action="store_true",
        help="Merge all secrets into one object. Secrets that are simple strings will have the last segment of their name (characters following the last /) as they key.",
    )
    parser.add_argument("secret", type=str, nargs="+")
    parser.add_argument(
        "--stdin",
        help=f"Read secret options from stdin. See details below for more info.\n{stdin_help_string}\n",
        default=(not stdin.isatty()),
    )
    casing_group = parser.add_mutually_exclusive_group()
    casing_group.add_argument(
        "-u",
        "--upper",
        action="store_true",
        help="Cast secret names to upper case. Does not apply to prefixes.",
    )
    casing_group.add_argument(
        "-l",
        "--lower",
        action="store_true",
        help="Cast secret names to lower case. Does not apply to prefixes.",
    )

    if args:
        res = parser.parse_args(args)
    else:
        res = parser.parse_args()

    return res


def parse_stdin() -> List[Namespace]:
    """Handles receiving special syntax on stdin for retreiving secrets"""
    arg_groups = []

    # Consume any args supplied through normal CLI args
    if len(argv) > 1:
        base_args = parse_args()

        # Create a separate entry in arg_groups so separate secret positional args
        # are treated the same way as separate entries from stdin
        for secret in base_args.secret:
            cmd = [secret]
            if base_args.upper:
                cmd.append("--upper")
            if base_args.lower:
                cmd.append("--lower")
            if base_args.merge_keys:
                cmd.append("--merge-keys")
            if base_args.prefix:
                cmd += ["--prefix", base_args.prefix]
            arg_groups.append(cmd)

    # Trim terminal garbage off of stdin
    stdin_args = [
        ANSI_RE.sub("", x).replace("\n", "").strip().split("|")
        for x in stdin.read().strip().split(";")
    ]

    # Parse each semicolon delimited entry into its own commandline args as if it were its own
    # invocation. --merge-keys will be overwritten by the main commandline inputs.
    for group in stdin_args:
        # When passing a list to `parse_args()` the positional arguments HAVE to come first.
        secret = group.pop(0).strip()
        cmd = [secret]
        if base_args.merge_keys:
            cmd.append("--merge")
        while group:
            next_cmd = group.pop(0).strip()
            if next_cmd.startswith("prefix"):
                cmd += ["--prefix", next_cmd.split(" ")[1]]
            else:
                cmd.append(f"--{next_cmd}")

        arg_groups.append(cmd)

    # Parse all of the inputs and get back regular argparse namespaces
    args = [parse_args(group) for group in arg_groups]
    return args


def main() -> None:
    args = parse_args()

    if args.stdin:
        args = parse_stdin()

    res = get_secrets(args)

    print(dumps(res, indent=2))


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        stderr.write("\rCanceled by user. Exiting.")
        exit(1)
